package com.loafle.overflow.central.module.infra.service;

import com.loafle.overflow.central.commons.utils.PageUtil;
import com.loafle.overflow.central.module.infra.dao.InfraDAO;
import com.loafle.overflow.central.module.infra.dao.InfraHostDAO;
import com.loafle.overflow.central.module.infra.dao.InfraServiceDAO;
import com.loafle.overflow.central.module.infra.dao.InfraZoneDAO;
import com.loafle.overflow.central.module.meta.service.CentralMetaCryptoTypeService;
import com.loafle.overflow.core.model.PageParams;
import com.loafle.overflow.model.probe.Probe;
import com.loafle.overflow.core.exception.OverflowException;
import com.loafle.overflow.model.discovery.Host;
import com.loafle.overflow.model.discovery.Port;
import com.loafle.overflow.model.discovery.Zone;
import com.loafle.overflow.model.infra.Infra;
import com.loafle.overflow.model.infra.InfraHost;
import com.loafle.overflow.model.infra.InfraHostIP;
import com.loafle.overflow.model.infra.InfraHostPort;
import com.loafle.overflow.model.infra.InfraZone;
import com.loafle.overflow.model.meta.MetaCryptoType;
import com.loafle.overflow.model.meta.MetaIPType;
import com.loafle.overflow.model.meta.MetaInfraType;
import com.loafle.overflow.model.meta.MetaPortType;
import com.loafle.overflow.model.meta.MetaTargetHostType;
import com.loafle.overflow.model.meta.MetaTargetZoneType;
import com.loafle.overflow.service.central.infra.InfraService;
import com.loafle.overflow.service.central.probe.ProbeService;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

@Service("InfraService")
public class CentralInfraService implements InfraService {
  @Autowired
  ProbeService probeService;

  @Autowired
  InfraDAO infraDAO;

  @Autowired
  InfraZoneDAO infraZoneDAO;

  @Autowired
  InfraHostDAO infraHostDAO;

  @Autowired
  InfraServiceDAO infraServiceDAO;

  @Autowired
  CentralInfraHostIPService infraHostIPService;

  @Autowired
  CentralInfraHostPortService infraHostPortService;

  @Autowired
  CentralMetaCryptoTypeService metaCryptoTypeService;

  @Override
  public Infra regist(Infra infra) throws OverflowException {
    return this.infraDAO.save(infra);
  }

  @Override
  @Transactional
  public InfraZone registZone(Long probeID, Zone zone) throws OverflowException {
    if (null == zone) {
      throw new OverflowException("Zone is not valid");
    }

    Probe probe = this.probeService.read(probeID);
    if (null == probe) {
      throw new OverflowException(String.format("ID[%d] of Probe is not valid", probeID));
    }

    if (null == zone.getNetwork()) {
      throw new OverflowException("Network is not valid");
    }

    if (null == zone.getIface()) {
      throw new OverflowException("Iface is not valid");
    }

    MetaIPType metaIPType = new MetaIPType();
    metaIPType.setKey(zone.getMetaIPTypeKey());

    InfraHostIP infraHostIP = new InfraHostIP();
    infraHostIP.setMetaIPType(metaIPType);
    infraHostIP.setAddress(zone.getAddress());
    infraHostIP.setMac(zone.getMac());

    infraHostIP = infraHostIPService.validate(infraHostIP);

    InfraZone infraZone = this.infraZoneDAO.findByProbeIdAndNetwork(probeID, zone.getNetwork());
    if (null != infraZone) {
      return infraZone;
    }

    infraZone = new InfraZone();
    infraZone.setMetaInfraType(MetaInfraType.Enum.ZONE.to());
    infraZone.setProbe(probe);
    infraZone.setMetaTargetZoneType(MetaTargetZoneType.Enum.ZONE.to());
    infraZone.setNetwork(zone.getNetwork());
    infraZone.setIface(zone.getIface());
    infraZone.setMetaIPType(infraHostIP.getMetaIPType());
    infraZone.setAddress(infraHostIP.getAddress());
    infraZone.setMac(infraHostIP.getMac());

    return this.infraZoneDAO.save(infraZone);
  }

  @Override
  @Transactional
  public InfraHost registHost(Long probeID, Host host) throws OverflowException {
    if (null == host) {
      throw new OverflowException("Host is not valid");
    }

    Probe probe = this.probeService.read(probeID);
    if (null == probe) {
      throw new OverflowException(String.format("ID[%d] of Probe is not valid", probeID));
    }

    Zone zone = host.getZone();
    if (null == zone) {
      throw new OverflowException("Zone is not valid");
    }

    InfraZone infraZone = this.registZone(probeID, zone);

    MetaIPType metaIPType = new MetaIPType();
    metaIPType.setKey(host.getMetaIPTypeKey());

    InfraHostIP infraHostIP = new InfraHostIP();
    infraHostIP.setMetaIPType(metaIPType);
    infraHostIP.setAddress(host.getAddress());
    infraHostIP.setMac(host.getMac());

    infraHostIP = infraHostIPService.validate(infraHostIP);
    metaIPType = infraHostIP.getMetaIPType();

    InfraHostIP eInfraHostIP = this.infraHostIPService.readByInfraHostInfraZoneIDAndMetaIPTypeKeyAndAddress(
        infraZone.getId(), metaIPType.getKey(), host.getAddress());
    if (null != eInfraHostIP) {
      throw new OverflowException(String.format("IP[%s(%s)] of Zone[%s] is exist already", host.getAddress(),
          metaIPType.getName(), infraZone.getNetwork()));
    }

    InfraHost infraHost = new InfraHost();
    infraZone.setMetaInfraType(MetaInfraType.Enum.HOST.to());
    infraZone.setProbe(probe);
    infraHost.setInfraZone(infraZone);
    infraHost.setMetaTargetHostType(MetaTargetHostType.Enum.UNKNOWN.to());
    infraHost.setInfraHostIPs(Arrays.asList(infraHostIP));

    return this.infraHostDAO.save(infraHost);
  }

  @Override
  @Transactional
  public com.loafle.overflow.model.infra.InfraService registService(Long probeID,
      com.loafle.overflow.model.discovery.Service service) throws OverflowException {

    if (null == service) {
      throw new OverflowException("Service is not valid");
    }

    if (null == service.getMetaCryptoTypeKey()) {
      throw new OverflowException("MetaCryptoTypeKey is not valid");
    }

    MetaCryptoType metaCryptoType = this.metaCryptoTypeService.readByKey(service.getMetaCryptoTypeKey());
    if (null == metaCryptoType) {
      throw new OverflowException(String.format("MetaCryptoTypeKey[%s] is not valid", service.getMetaCryptoTypeKey()));
    }

    Probe probe = this.probeService.read(probeID);
    if (null == probe) {
      throw new OverflowException(String.format("ID[%d] of Probe is not valid", probeID));
    }

    Port port = service.getPort();
    if (null == port) {
      throw new OverflowException("Port of Service is not valid");
    }

    Host host = port.getHost();
    if (null == host) {
      throw new OverflowException("Host of Service is not valid");
    }

    Zone zone = host.getZone();
    if (null == zone) {
      throw new OverflowException("Zone is not valid");
    }

    MetaIPType metaIPType = new MetaIPType();
    metaIPType.setKey(host.getMetaIPTypeKey());

    InfraHostIP infraHostIP = new InfraHostIP();
    infraHostIP.setMetaIPType(metaIPType);
    infraHostIP.setAddress(host.getAddress());
    infraHostIP.setMac(host.getMac());

    infraHostIP = this.infraHostIPService.validate(infraHostIP);
    metaIPType = infraHostIP.getMetaIPType();

    MetaPortType metaPortType = new MetaPortType();
    metaPortType.setKey(port.getMetaPortTypeKey());

    InfraHostPort infraHostPort = new InfraHostPort();
    infraHostPort.setMetaPortType(metaPortType);
    infraHostPort.setPort(port.getPortNumber());

    infraHostPort = this.infraHostPortService.validate(infraHostPort);
    metaPortType = infraHostPort.getMetaPortType();

    InfraZone infraZone = this.registZone(probeID, zone);
    InfraHost infraHost = this.registHost(probeID, host);

    infraHostIP = this.infraHostIPService.readByInfraHostIDAndMetaIPTypeKeyAndAddress(infraHost.getId(),
        metaIPType.getKey(), infraHostIP.getAddress());
    if (null == infraHostIP) {
      throw new OverflowException(String.format("IP[%s(%s)] of Zone[%s] is not exist", host.getAddress(),
          metaIPType.getName(), infraZone.getNetwork()));
    }

    InfraHostPort eInfraHostPort = this.infraHostPortService
        .readByInfraHostIDAndMetaPortTypeKeyAndPort(infraHost.getId(), metaPortType.getKey(), infraHostPort.getPort());
    if (null != eInfraHostPort) {
      throw new OverflowException(
          String.format("Port[%d(%s)] of Zone[%s].Host[%s(%s)] is not exist", host.getAddress(), metaPortType.getName(),
              infraZone.getNetwork(), infraHostIP.getAddress(), infraHostIP.getMetaIPType().getName()));
    }

    infraHostPort = this.infraHostPortService.regist(infraHostPort);

    com.loafle.overflow.model.infra.InfraService infraService = new com.loafle.overflow.model.infra.InfraService();
    infraService.setInfraHostIP(infraHostIP);
    infraService.setInfraHostPort(infraHostPort);
    infraService.setMetaCryptoType(metaCryptoType);

    return this.infraServiceDAO.save(infraService);
  }

  @Override
  @Transactional
  public List<Infra> registDiscoverd(Long probeID, List<Host> hosts,
      List<com.loafle.overflow.model.discovery.Service> services) throws OverflowException {

    List<Infra> infras = new ArrayList<>();

    if (null != hosts) {
      for (Host host : hosts) {
        infras.add(this.registHost(probeID, host));
      }
    }
    if (null != services) {
      for (com.loafle.overflow.model.discovery.Service service : services) {
        infras.add(this.registService(probeID, service));
      }
    }

    return infras;
  }

  public InfraHost registByNoAuthProbe(Long probeID, InfraHost infraHost) throws OverflowException {
    // NoAuthProbeDescriptionNetwork networkInfo =
    // noAuthProbeDescription.getNetwork();
    // InfraHostIP infraHostIP = new InfraHostIP();
    // // FIXME getting IPType from networkInfo
    // infraHostIP.setMetaIPType(new MetaIPType((short) 1));
    // infraHostIP.setAddress(networkInfo.getAddress());
    // infraHostIP.setMac(networkInfo.getMacAddress());
    // List<InfraHostIP> infraHostIPs = new ArrayList<>();
    // infraHostIPs.add(infraHostIP);

    // InfraHost infraHost = new InfraHost();
    // infraHost.setProbe(new Probe(probeID));
    // infraHost.setMetaInfraType(MetaInfraType.Enum.HOST.to());
    // infraHost.setMetaTargetHostType(new MetaTargetHostType(1));
    // infraHost.setInfraHostIPs(infraHostIPs);

    // return this.infraDAO.save(infraHost);

    return null;
  }

  @Override
  public Infra read(Long id) throws OverflowException {
    return this.infraDAO.findById(id).get();
  }

  @Override
  public Page<Infra> readAllByProbeID(Long probeID, PageParams pageParams) throws OverflowException {
    return this.infraDAO.findAllByProbeId(probeID, PageUtil.getPageRequest(pageParams));
  }

  @Override
  public Page<Infra> readAllByDomainID(Long domainID, PageParams pageParams) throws OverflowException {
    return this.infraDAO.findAllByProbeDomainId(domainID, PageUtil.getPageRequest(pageParams));
  }

  @Override
  public List<Infra> readAllInfraByProbeID(Long probeID) throws OverflowException {
    return this.infraDAO.findAllByProbeId(probeID);
  }

  @Override
  public List<Infra> readAllInfraByDomainID(Long domainID) throws OverflowException {
    return this.infraDAO.findAllByProbeDomainId(domainID);
  }

  public List<InfraZone> readAllInfraZoneByProbeDomainID(Long probeDomainID) throws OverflowException {
    return null;
  }


}
