package com.loafle.overflow.central.module.noauthprobe.service;

import com.loafle.overflow.central.commons.service.MessagePublisher;
import com.loafle.overflow.core.annotation.ProbeAPI;
import com.loafle.overflow.core.annotation.WebappAPI;
import com.loafle.overflow.central.commons.utils.GenerateKey;
import com.loafle.overflow.central.commons.utils.SessionMetadata;
import com.loafle.overflow.central.module.infra.service.CentralInfraService;
import com.loafle.overflow.central.module.meta.service.CentralMetaIPTypeService;
import com.loafle.overflow.central.module.meta.service.CentralMetaTargetHostTypeService;
import com.loafle.overflow.central.module.noauthprobe.dao.NoAuthProbeDAO;
import com.loafle.overflow.core.exception.OverflowException;
import com.loafle.overflow.model.apikey.ApiKey;
import com.loafle.overflow.model.domain.Domain;
import com.loafle.overflow.model.domain.DomainMember;
import com.loafle.overflow.model.infra.InfraHost;
import com.loafle.overflow.model.infra.InfraHostIP;
import com.loafle.overflow.model.infra.InfraHostOS;
import com.loafle.overflow.model.infra.InfraZone;
import com.loafle.overflow.model.member.Member;
import com.loafle.overflow.model.meta.MetaIPType;
import com.loafle.overflow.model.meta.MetaInfraType;
import com.loafle.overflow.model.meta.MetaNoAuthProbeStatus;
import com.loafle.overflow.model.meta.MetaProbeStatus;
import com.loafle.overflow.model.meta.MetaTargetHostType;
import com.loafle.overflow.model.noauthprobe.NoAuthProbe;
import com.loafle.overflow.model.probe.Probe;
import com.loafle.overflow.model.probe.ProbeHost;
import com.loafle.overflow.service.central.apikey.ApiKeyService;
import com.loafle.overflow.service.central.domain.DomainMemberService;
import com.loafle.overflow.service.central.meta.MetaIPTypeService;
import com.loafle.overflow.service.central.noauthprobe.NoAuthProbeService;
import com.loafle.overflow.service.central.probe.ProbeHostService;
import com.loafle.overflow.service.central.probe.ProbeService;

import org.codehaus.jackson.map.ObjectMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.stereotype.Service;

import inet.ipaddr.IPAddress;
import inet.ipaddr.IPAddressString;

import java.io.IOException;
import java.util.Date;
import java.util.List;
import java.util.UUID;

import javax.transaction.Transactional;

/**
 * Created by snoop on 17. 6. 28.
 */
@Service("NoAuthProbeService")
public class CentralNoAuthProbeService implements NoAuthProbeService {

  @Autowired
  private NoAuthProbeDAO noAuthProbeDAO;

  @Autowired
  private ApiKeyService apiKeyService;

  @Autowired
  private ProbeService probeService;

  @Autowired
  private MessagePublisher messagePublisher;

  @Autowired
  private DomainMemberService domainMemberService;

  @Autowired
  private ObjectMapper objectMapper;

  @Autowired
  private CentralInfraService infraService;

  @Autowired
  private ProbeHostService probeHostService;

  @Autowired
  private CentralMetaIPTypeService metaIPTypeService;

  @Autowired
  private CentralMetaTargetHostTypeService metaTargetHostTypeService;

  @ProbeAPI
  public NoAuthProbe regist(NoAuthProbe noAuthProbe) throws OverflowException {
    if (null == noAuthProbe) {
      throw new OverflowException("NoAuthProbe is not valid");
    }

    ApiKey apiKey = apiKeyService.readByApiKey(noAuthProbe.getApiKey());
    if (null == apiKey) {
      throw new OverflowException(String.format("APIKey[%s] of NoAuthProbe is not valid", noAuthProbe.getApiKey()));
    }

    noAuthProbe.setDomain(apiKey.getDomain());
    noAuthProbe.setTempProbeKey(GenerateKey.getKey());
    noAuthProbe.setMetaNoAuthProbeStatus(MetaNoAuthProbeStatus.Enum.PROCESSING.to());

    messagePublisher.publishToDomainMembers(apiKey.getDomain().getId(), "NoAuthProbeService.regist", noAuthProbe);

    return this.noAuthProbeDAO.save(noAuthProbe);
  }

  public List<NoAuthProbe> readAllByDomainID(Long domainID) throws OverflowException {
    return this.noAuthProbeDAO.findAllByDomainIdAndMetaNoAuthProbeStatusId(domainID,
        MetaNoAuthProbeStatus.Enum.PROCESSING.getValue());
  }

  public NoAuthProbe read(Long id) {
    return this.noAuthProbeDAO.findById(id).get();
  }

  @WebappAPI
  @Transactional
  public List<NoAuthProbe> acceptNoAuthProbe(Long noAuthProbeID, String zoneCIDR) throws OverflowException {
    NoAuthProbe noAuthProbe = this.noAuthProbeDAO.findById(noAuthProbeID).get();
    if (null == noAuthProbe) {
      throw new OverflowException(String.format("ID(%s) of NoAuthProbe is not exist", noAuthProbeID));
    }

    InfraHost oriInfraHost = null;
    try {
      oriInfraHost = this.objectMapper.readValue(noAuthProbe.getInfraHostMeta(), InfraHost.class);
    } catch (IOException e) {
      throw new OverflowException("json error", e);
    }

    if (null == oriInfraHost.getInfraHostIPs() || 0 == oriInfraHost.getInfraHostIPs().size()) {
      throw new OverflowException("IP of Probe InfraHost is not valild");
    }

    if (null == zoneCIDR && 1 < oriInfraHost.getInfraHostIPs().size()) {
      throw new OverflowException(String.format("Zone count[%d] of Probe is not single, must specified",
          oriInfraHost.getInfraHostIPs().size()));
    }

    InfraHostIP infraHostIP = null;
    if (1 == oriInfraHost.getInfraHostIPs().size()) {
      infraHostIP = oriInfraHost.getInfraHostIPs().get(0);
    } else {
      IPAddress ZoneIPAddress = new IPAddressString(zoneCIDR).getAddress();

      for (InfraHostIP _infraHostIP : oriInfraHost.getInfraHostIPs()) {
        if (null == _infraHostIP.getAddress()) {
          continue;
        }
        IPAddress ipAddress = new IPAddressString(_infraHostIP.getAddress()).getAddress();
        if (ipAddress.toPrefixBlock().equals(ZoneIPAddress)) {
          infraHostIP = _infraHostIP;
          break;
        }
      }
    }
    if (null == infraHostIP) {
      throw new OverflowException("Network is not valid");
    }

    if (null == oriInfraHost.getInfraHostOS()) {
      throw new OverflowException("InfraHostOS is not valid");
    }
    if (null == oriInfraHost.getInfraHostMachine()) {
      throw new OverflowException("InfraHostMachine is not valid");
    }

    BCryptPasswordEncoder passwordEncoder = new BCryptPasswordEncoder();
    String encryptKey = passwordEncoder.encode(UUID.randomUUID().toString());

    ApiKey apiKey = apiKeyService.readByApiKey(noAuthProbe.getApiKey());
    String memberEmail = SessionMetadata.getTargetID();
    DomainMember domainMember = domainMemberService.readByMemberEmail(memberEmail);

    Probe probe = new Probe();
    probe.setEncryptionKey(encryptKey);
    probe.setProbeKey(GenerateKey.getKey());
    probe.setDomain(new Domain(apiKey.getDomain().getId()));
    probe.setAuthorizeMember(new Member(domainMember.getMember().getId()));
    probe.setTargetCount(0);
    probe.setMetaProbeStatus(MetaProbeStatus.Enum.INITIAL.to());

    String name = oriInfraHost.getInfraHostOS().getName();
    name += " probe";
    probe.setName(name);
    IPAddress ipAddress = new IPAddressString(infraHostIP.getAddress()).getAddress();
    probe.setCidr(ipAddress.toPrefixBlock().toString());

    probe = this.probeService.regist(probe);

    InfraZone infraZone = this.newInfraZone(infraHostIP);

    InfraHost infraHost = this.newInfraHost(oriInfraHost);
    infraHost.setInfraZone(infraZone);

    Probe probe = this.newProbe(noAuthProbe, infraZone, infraHost);
    infraZone.setProbe(probe);
    infraHost.setProbe(probe);

    infraZone = this.infraService.registInfraZone(infraZone);
    infraHost = this.infraService.registInfraHost(infraHost);

    this.newProbeHost(probe, infraHost);

    noAuthProbe.setMetaNoAuthProbeStatus(MetaNoAuthProbeStatus.Enum.ACCEPTED.to());
    this.noAuthProbeDAO.save(noAuthProbe);

    messagePublisher.publishToNoAuthProbe(noAuthProbe.getTempProbeKey(), "NoAuthProbeService.Accept",
        probe.getProbeKey());

    return this.readAllByDomainID(noAuthProbe.getDomain().getId());
  }

  private InfraZone newInfraZone(InfraHostIP infraHostIP) throws OverflowException {
    if (null == infraHostIP) {
      throw new OverflowException("InfraHostIP is not valid");
    }

    InfraZone infraZone = new InfraZone();
    infraZone.setMetaInfraType(MetaInfraType.Enum.ZONE.to());

    if (null == infraHostIP.getMetaIPType() || null == infraHostIP.getMetaIPType().getKey()) {
      throw new OverflowException("MetaIPType is not valid");
    }
    MetaIPType metaIPType = this.metaIPTypeService.readByKey(infraHostIP.getMetaIPType().getKey());
    if (null == metaIPType) {
      throw new OverflowException(
          String.format("Key[%s] of MetaIPType is not valid", infraHostIP.getMetaIPType().getKey()));
    }
    infraZone.setMetaIPType(metaIPType);

    if (null == infraHostIP.getAddress()) {
      throw new OverflowException("Address is not valid");
    }
    infraZone.setAddress(infraHostIP.getAddress());

    IPAddress ipAddress = new IPAddressString(infraHostIP.getAddress()).getAddress();
    infraZone.setNetwork(ipAddress.toPrefixBlock().toString());

    if (null == infraHostIP.getAddress()) {
      throw new OverflowException("Iface is not valid");
    }
    infraZone.setIface(infraHostIP.getIface());

    if (null == infraHostIP.getMac()) {
      throw new OverflowException("Mac is not valid");
    }
    infraZone.setMac(infraHostIP.getMac());

    return infraZone;
  }

  private InfraHost newInfraHost(InfraHost oriInfraHost) throws OverflowException {
    InfraHost infraHost = new InfraHost();
    infraHost.setMetaInfraType(MetaInfraType.Enum.HOST.to());

    InfraHostOS infraHostOS = oriInfraHost.getInfraHostOS();
    // MetaTargetHostType metaTargetHostType = this.metaTargetHostTypeService.readByInfraHostOS(infraHostOS);
    infraHost.setMetaTargetHostType(MetaTargetHostType.Enum.UNKNOWN.to());
    infraHost.setInfraHostOS(infraHostOS);

    infraHost.setInfraHostIPs(oriInfraHost.getInfraHostIPs());
    infraHost.setInfraHostPorts(oriInfraHost.getInfraHostPorts());
    infraHost.setInfraHostApplications(oriInfraHost.getInfraHostApplications());
    infraHost.setInfraHostDaemons(oriInfraHost.getInfraHostDaemons());

    return infraHost;
  }

  private Probe newProbe(NoAuthProbe noauthprobe, InfraZone infraZone, InfraHost infraHost) throws OverflowException {

    BCryptPasswordEncoder passwordEncoder = new BCryptPasswordEncoder();
    String encryptKey = passwordEncoder.encode(UUID.randomUUID().toString());

    ApiKey apiKey = apiKeyService.readByApiKey(noauthprobe.getApiKey());
    String memberEmail = SessionMetadata.getTargetID();
    DomainMember domainMember = domainMemberService.readByMemberEmail(memberEmail);

    Probe probe = new Probe();
    probe.setEncryptionKey(encryptKey);
    probe.setProbeKey(GenerateKey.getKey());
    probe.setDomain(new Domain(apiKey.getDomain().getId()));
    probe.setAuthorizeMember(new Member(domainMember.getMember().getId()));
    probe.setTargetCount(0);
    probe.setMetaProbeStatus(MetaProbeStatus.Enum.INITIAL.to());

    String name = infraHost.getInfraHostOS().getName();
    name += " probe";
    probe.setName(name);
    probe.setCidr(infraZone.getNetwork());

    return this.probeService.regist(probe);
  }

  private void newProbeHost(Probe probe, InfraHost infraHost) throws OverflowException {
    ProbeHost probeHost = new ProbeHost();
    probeHost.setInfraHost(infraHost);
    probeHost.setProbe(probe);

    this.probeHostService.regist(probeHost);
  }

  @WebappAPI
  public List<NoAuthProbe> denyNoauthProbe(Long noAuthProbeID) throws OverflowException {
    NoAuthProbe noAuthProbe = this.noAuthProbeDAO.findById(noAuthProbeID).get();

    noAuthProbe.setMetaNoAuthProbeStatus(MetaNoAuthProbeStatus.Enum.DENIED.to());
    this.noAuthProbeDAO.save(noAuthProbe);

    messagePublisher.publishToNoAuthProbe(noAuthProbe.getTempProbeKey(), "NoAuthProbeService.Deny");

    return this.readAllByDomainID(noAuthProbe.getDomain().getId());
  }

  public NoAuthProbe readByTempProbeKey(String tempKey) throws OverflowException {
    return this.noAuthProbeDAO.findByTempProbeKey(tempKey);
  }

  public void onConnect(String tempKey, String connectAddress) throws OverflowException {
    NoAuthProbe noAuthProbe = this.noAuthProbeDAO.findByTempProbeKey(tempKey);
    noAuthProbe.setConnectDate(new Date());
    noAuthProbe.setConnectAddress(connectAddress);
    noAuthProbe = this.noAuthProbeDAO.save(noAuthProbe);

    messagePublisher.publishToDomainMembers(noAuthProbe.getDomain().getId(), "NoAuthProbeService.onConnect",
        noAuthProbe);
  }

  public void onDisconnect(String tempKey) throws OverflowException {
    NoAuthProbe noAuthProbe = this.noAuthProbeDAO.findByTempProbeKey(tempKey);
    noAuthProbe.setConnectDate(null);
    noAuthProbe.setConnectAddress(null);
    noAuthProbe = this.noAuthProbeDAO.save(noAuthProbe);

    messagePublisher.publishToDomainMembers(noAuthProbe.getDomain().getId(), "NoAuthProbeService.onDisconnect",
        noAuthProbe);
  }

}
